#!/usr/bin/env python
# -*- coding: UTF8 -*-
"""https://oxfordstrat.com/coasdfASD32/uploads/2016/01/turtle-rules.pdf

Exits
The Turtles used breakout based exits for profitable positions.

There is an old saying: "you can never go broke taking a profit." The 
Turtles would not agree with this statement. Getting out of winning positions
too early, i.e. "taking a profit" too early, is one of the most common mistakes
when trading trend following systems.

Prices never go straight up; therefore it is necessary to let the prices go against you if
you are going to ride a trend. Early in a trend this can often mean watching decent
profits of 10% to 30% fade to a small loss. In the middle of a trend, it might mean
watching a profit of 80% to 100% drop by 30% to 40%. The temptation to lighten the 
position to "lock in profits" can be very great.

The Turtles knew that where you took a profit could make the difference between
winning and losing.

The Turtle System enters on breakouts. Most breakouts do not result in trends. This
means that most of the trades that the Turtles made resulted in losses. If the winning
trades did not earn enough on average to offset these losses, the Turtles would have
lost money. Every profitable trading system has a different optimal exit point.

Consider the Turtle System; if you exit winning positions at a 1 N profit while you
exited losing positions at a 2 N loss you would need twice as many winners to offset
the losses from the losing trades.

There is a complex relationship between the components of a trading system. This
means that you can't consider the proper exit for a profitable position without
considering the entry, money management and other factors.

The proper exit for winning positions is one of the most important aspects of trading,
and the least appreciated. Yet it can make the difference between winning and losing.

Turtle Exits

The System 1 exit was a 10 day low for long positions and a 10 day high for short
positions. All the Units in the position would be exited if the price went against the
position for a 10 day breakout.

The System 2 exit was a 20 day low for long positions and a 20 day high for short
positions. All the Units in the position would be exited if the price went against the
position for a 20 day breakout.

As with entries, the Turtles did not typically place exit stop orders, but instead watched
the price during the day, and started to phone in exit orders as soon as the price traded
through the exit breakout price.
"""

# n
# breakout
# long or short: bool

def getexit(n, breakout, type: bool):
  """Return a decision on whether to exit or not.

  Args:
    n: 'N' of the breakout.
    breakout: The breakout price, i.e. the first unit entered.
    units: How many units in that same position there are.
  """
  return Exit(n=n, breakout=breakout, type=type)


class Exit:
  def __init__(self, n, breakout, type: bool):

    self.exit_signal = self.signal()
  
  

  def signal(self) -> bool:
    """Return a boolean determing whether to exit or not.

    Args:
      

    Returns:
      True to exit, False to not. 
    """
    # First, get the  


#!/usr/bin/env python
# -*- coding: UTF8 -*-
"""https://oxfordstrat.com/coasdfASD32/uploads/2016/01/turtle-rules.pdf

System 1 - A shorter-term system based on a 20-day breakout.

System 2 - A simpler long-term system based on a 55-day breakout.

Breakouts
A breakout is defined as the price exceeding the high or low of a particular number of
days. Thus a 20-day breakout would be defined as exceeding the high or low of the
preceding 20 days.

Turtles always traded at the breakout when it was exceeded during the day, and did not
wait until the daily close or the open of the following day. In the case of opening gaps,
the Turtles would enter positions on the open if a market opened through the price of
the breakout.

System 1 Entry - Turtles entered positions when the price exceeded by a single tick the
high or low of the preceding 20 days. If the price exceeded the 20-day high, then the
Turtles would buy one Unit to initiate a long position in the corresponding
commodity. If the price dropped one tick below the low of the last 20-days, the Turtles
would sell one Unit to initiate a short position.

System 1 breakout entry signals would be ignored if the last breakout would have
resulted in a winning trade. NOTE: For the purposes of this test, the last breakout was
considered the last breakout in the particular commodity irrespective of whether or not
that particular breakout was actually taken, or was skipped because of this rule. This
breakout would be considered a losing breakout if the price subsequent to the date of
the breakout moved 2N against the position before a profitable 10-day exit occurred.

The direction of the last breakout was irrelevant to this rule. Thus, a losing long
breakout or a losing short breakout would enable the subsequent new breakout to be
taken as a valid entry, regardless of its direction (long or short).

However, in the event that a System 1 entry breakout was skipped because the
previous trade had been a winner, an entry would be made at the 55-day breakout to
avoid missing major moves. This 55-day breakout was considered the Failsafe
Breakout point.

At any given point, if you were out of the market, there would always be some price
which would trigger a short entry and another different and higher price which would
trigger a long entry. If the last breakout was a loser, then the entry signal would be
closer to the current price (i.e. the 20 day breakout), than if it had been a winner, in
which case the entry signal would likely be farther away, at the 55 day breakout.

System 2 Entry - Entered when the price exceeded by a single tick the high or low of
the preceding 55 days. If the price exceeded the 55 day high, then the Turtles would
buy one Unit to initiate a long position in the corresponding commodity. If the price
dropped one tick below the low of the last 55 days, the Turtles would sell one Unit to
initiate a short position.

All breakouts for System 2 would be taken whether the previous breakout had been a
winner or not.
"""
import sys
import pandas as pd
from pathlib import Path

""" Avoid import errors while importing 'getn' function """
path = str(Path(Path(__file__).parent.absolute()).parent.absolute())
sys.path.insert(0, path)
from position_sizing.position_sizing import getn


def getsignal(price: float, dataframe: pd.DataFrame, system: int) -> bool:
  """Get an entry signal.

  COMMENT(s): 
    Consider removing dataframe arg and doing that inside function.
    Move from function to class?

  Args:
    price: The asset's current price.
    dataframe: The asset's get_data() dataframe.
    system: Which system to use, 1 or 2.
  """
  if system not in (1, 2):
    raise AssertionError("System must be of of 1 or 2.")
  
  # reverse the dataframe
  dataframe = dataframe.loc[::-1]

  if system == 1:
    return system_one(price=price, dataframe=dataframe)
  elif system == 2:
    return system_two(price=price, dataframe=dataframe) 


def breakout(price: float, dataframe: pd.DataFrame, days: int) -> bool:
  """Determine if the price is a breakout price.
  
  A breakout is defined as the price exceeding the high or low of a particular number of
  days. Thus a 20-day breakout would be defined as exceeding the high or low of the
  preceding 20 days.

  Turtles always traded at the breakout when it was exceeded during the day, and did not
  wait until the daily close or the open of the following day. In the case of opening gaps,
  the Turtles would enter positions on the open if a market opened through the price of
  the breakout.

  Args: 
    price: The asset's current price.
    dataframe: The asset's get_data() dataframe.
    days: The number of days.
    
  Returns:
    True if long breakout, False if short breakout, None if no breakout.
  """
  if all(_ <= price for _ in dataframe.head(days+1)["high"].tolist()): # if all the highs are lower than the price
    return True # long
  
  elif all(_ >= price for _ in dataframe.head(days+1)["low"].tolist()): # if all the lows are higher than the price
    return False # short

  return None # no breakout
  

def last_breakout(dataframe: pd.DataFrame, stand_devs: float = 2.0, days: int = 20, skip: int = 1):
  """Get the last breakout and determine if it was a losing or winning position.

  System 1 breakout entry signals would be ignored if the last breakout would have
  resulted in a winning trade. NOTE: For the purposes of this test, the last breakout was
  considered the last breakout in the particular commodity irrespective of whether or not 
  that particular breakout was actually taken, or was skipped because of this rule. This
  breakout would be considered a losing breakout if the price subsequent to the date of
  the breakout moved 2N against the position before a profitable 10-day exit occurred.

  Args:
    dataframe: An asset's full dataframe.
    n: 'N' of the asset at the price. 
    stand_devs: The number of standard deviations.
    days: The number of days.
    skip: How many breakouts to skip. 

  Returns:
    True if the last breakout was a winning position, False if losing.
  """
  # 1: Find the last breakout.
  loc: int # the location of the last breakout
  direction: bool # long or short 
  n_dataframe: pd.DataFrame

  for index, row in enumerate(dataframe.iterrows()):
    last_twenty = dataframe.iloc[index+1: index+days+1] # the last twenty days
    _breakout = breakout(price=row[1]["close"], dataframe=last_twenty, days=20) # is the first row a breakout
 
    if isinstance(_breakout, bool) and index > skip: # if a breakout
      loc, direction, n_dataframe = index, True if _breakout is True else False, last_twenty
      break

  # 2: Find 'N' of the last breakout.
  N = getn(asset="", dataframe=n_dataframe)
  
  # 3: Determine if winning or losing position.
  breakout_price = dataframe.iloc[loc]["close"] # the last breakout price

  for index, row in enumerate(dataframe.iloc[:loc+1].iloc[::-1].iterrows()):  
    # if the difference between the close and breakout price is greater than num * n, return False
    if index == 10: break

    if direction: # if a long position, True
      if row[1]["close"] - breakout_price >= stand_devs * N:
        return False

    if not direction: # if a short position, False
      if row[1]["close"] - breakout_price <= -(stand_devs * N):
        return False
  
  # if False not returned, position is winning.. return True
  return True


def system_one(price: float, dataframe: pd.DataFrame) -> bool or None or tuple[bool, int]:
  """System One Entry

  Turtles entered positions when the price exceeded by a single tick the
  high or low of the preceding 20 days. If the price exceeded the 20-day high, then the
  Turtles would buy one Unit to initiate a long position in the corresponding
  commodity. If the price dropped one tick below the low of the last 20-days, the Turtles
  would sell one Unit to initiate a short position.

  System 1 breakout entry signals would be ignored if the last breakout would have
  resulted in a winning trade. NOTE: For the purposes of this test, the last breakout was
  considered the last breakout in the particular commodity irrespective of whether or not
  that particular breakout was actually taken, or was skipped because of this rule. This
  breakout would be considered a losing breakout if the price subsequent to the date of
  the breakout moved 2N against the position before a profitable 10-day exit occurred.

  The direction of the last breakout was irrelevant to this rule. Thus, a losing long
  breakout or a losing short breakout would enable the subsequent new breakout to be
  taken as a valid entry, regardless of its direction (long or short).

  However, in the event that a System 1 entry breakout was skipped because the
  previous trade had been a winner, an entry would be made at the 55-day breakout to
  avoid missing major moves. This 55-day breakout was considered the Failsafe
  Breakout point.

  Args:
    price: The asset's current price.
    dataframe: The asset's full dataframe.

  Returns: 
    True if long breakout, False if short breakout, None if no breakout.
  """
  _breakout = breakout(price=price, dataframe=dataframe, days=20) # is the current price a 20-day breakout
  if isinstance(_breakout, bool): # if a breakout...
    previous_breakout = last_breakout(dataframe=dataframe, stand_devs=2.00) # get last breakout
    if previous_breakout: # if last breakout was losing, continue 
      return _breakout # true or false
    elif not previous_breakout: # if last breakout was winning, ignore
      if isinstance(breakout(price=price, dataframe=dataframe, days=55), bool): # if this breakout is a 55 day breakout
        return (_breakout, 55)


def system_two(price: float, dataframe: pd.DataFrame) -> bool or None:
  """System Two Entry

  Entered when the price exceeded by a single tick the high or low of
  the preceding 55 days. If the price exceeded the 55 day high, then the Turtles would
  buy one Unit to initiate a long position in the corresponding commodity. If the price
  dropped one tick below the low of the last 55 days, the Turtles would sell one Unit to
  initiate a short position.

  All breakouts for System 2 would be taken whether the previous breakout had been a
  winner or not.

  Args:
    price: The asset's current price.
    dataframe: The asset's full dataframe.

  Returns:
    True if long breakout, False if short breakout, None if no breakout.
  """
  _breakout = breakout(price=price, dataframe=dataframe, days=55) # is the current price is a 55-day breakout
  if isinstance(_breakout, bool): # if a breakout...
    return _breakout


# #!/usr/bin/env python
# -*- coding: UTF8 -*-
"""https://oxfordstrat.com/coasdfASD32/uploads/2016/01/turtle-rules.pdf

Adding Units
Turtles entered single Unit long positions at the breakouts and added to those
positions at 1/2 N intervals following their initial entry. This 1/2 N interval was based on 
the actual fill price of the previous order. So if an initial breakout order slipped by 1/2 N, 
then the new order would be 1 full N past the breakout to account for the 1/2 N
slippage, plus the normal 1/2 N unit add interval.

This would continue right up to the maximum permitted number of units. If the
market moved quickly enough it was possible to add the maximum four Units in a 
single day.

Example: 
Gold
N = 2.50
55 day breakout = 310

First Unit added    310.00
Second Unit         310.00 + 1/2 (2.50) or 311.25
Third Unit          311.25 + 1/2 (2.50) or 312.50
Fourth Unit         312.50 + 1/2 (2.50) or 313.75

Crude Oil
N = 1.20
55 day breakout = 28.30

First Unit added    28.30
Second Unit         28.30 + 1/2 (1.20) or 28.90
Third Unit          28.90 + 1/2 (1.20) or 29.50
Fourth Unit         29.50 + 1/2 (1.20) or 30.10
"""

def addunits(N: float, breakout: float, units: int = 4):
  """Returns the next four units.

  COMMENT: The maximum units allowed by the Turtle Trading system on one position is 4.

  Args:
    N: The 'N' at the breakout price.
    breakout: The breakout price.
    units: Optional, how many units to add. 
  """
  _units = [breakout]
  for i in range(units-1):
    _units.append(_units[i] + 0.5 * (N))

  return _units


#!/usr/bin/env python
# -*- coding: UTF8 -*-
"""https://oxfordstrat.com/coasdfASD32/uploads/2016/01/turtle-rules.pdf

Stop Placement
The Turtles placed their stops based on position risk. No trade could incur more than
2% risk.

Since 1 N of price movement represented 1% of Account Equity, the maximum stop
that would allow 2% risk would be 2 N of price movement. Turtle stops were set at 2
N below the entry for long positions, and 2 N above the entry for short positions.

In order to keep total position risk at a minimum, if additional units were added, the
stops for earlier units were raised by ½ N. This generally meant that all the stops for
the entire position would be placed at 2 N from the most recently added unit.
However, in cases where later units were placed at larger spacing either because of fast
markets causing skid, or because of opening gaps, there would be differences in the
stops.

For example:
Crude Oil
N = 1.20
55 day breakout = 28.30

              Entry Price   Stop
First Unit    28.30         25.90

              Entry Price   Stop
First Unit    28.30         26.50
Second Unit   28.90         26.50

              Entry Price   Stop
First Unit    28.30         27.10
Second Unit   28.90         27.10
Third Unit    29.50         27.10

              Entry Price   Stop
First Unit    28.30         27.70
Second Unit   28.90         27.70
Third Unit    29.50         27.70
Fourth Unit   30.10         27.70

Case where fourth unit was added at a higher price because the market opened gapping
up to 30.80:
              Entry Price   Stop
First Unit    28.30         27.70
Second Unit   28.90         27.70
Third Unit    29.50         27.70
Fourth Unit   30.80         28.40
"""
import pandas as pd

VERSIONS = ["reg", "whipsaw"]



def getstop(n: float,  entries: list[float], stand_devs: float = 2.00, mode: str = "reg"):
  """Get the Stop Prices.

  Args: 
    n: The N of the first entry.
    entries: A list of entry prices. Use addunits function from entries module.
    mode: The regular stop strategy or the whipsaw.
  """
  return Stop(n=n, entries=entries, stand_devs=stand_devs, mode=mode)


class Stop:
  """Represents units and their corresponding stops.

  Args:
    n:
    entries: A list of entry prices. Expected that they are in order.
    mode:
  """
  def __init__(self, n: float, entries: list[float], stand_devs: float = 2.00, mode: str = "reg"):
    risk = stand_devs * n

    if mode == "reg":
      self.stop = self.stop_placement(entries=entries, risk=risk, n=n)

  def stop_placement(self, entries: list[float], risk: float, n: float):
    """

    Args:
      entries:
      risk: 
    """
    diff = round(.5 * n, 2)
    print(diff)

    # the differences between all the entries
    check = [round(elem, 2) for elem in [j-i for i, j in zip(entries[:-1], entries[1:])]] 
    print(check)

    if all(i == diff for i in check): # if no gaps
      num_of_units = len(entries)
      stop = entries[num_of_units-1] - risk
      return stop
    
    else: # if a gap(s)
      
      # find gapped numbers
      gapped_entries = [i for i in check if i != diff]
      # print(gapped_entries)
      
      # get the indices for all gapped numbers
      indices: list = []
      for i in gapped_entries:
        ele = check.index(i)+1
        indices.append(ele)
      
      # now, get the stops
      stops: list = []
      for index, entry in enumerate(entries):
        if index in indices:
          pass

        # fucking this is annoying, I don't wanna do it rn moving on to exits
        



      





     
    

"""
Alternate Stop Strategy - The Whipsaw

The Turtles were told of an alternate stop strategy that resulted in better profitability,
but that was harder to execute because it incurred many more losses, which resulted in 
a lower win/loss ratio. This strategy was called the Whipsaw.

Instead of taking a 2% risk on each trade, the stops were placed at 1/2 N for 1/2%
account risk. If a given Unit was stopped out, the Unit would be re-entered if the
market reached the original entry price. A few Turtles traded this method with good
success.

The Whipsaw also had the added benefit of not requiring the movement of stops for
earlier Units as new Units were added, since the total risk would never exced 2% at 
the maximum four Units.

For example, using Whipsaw stops, the Crude Oil entry stops would be:

Crude Oil
N = 1.20
55 day breakout = 28.30

              Entry Price   Stop
First Unit    28.30         27.70

              Entry Price   Stop
First Unit    28.30         27.70
Second Unit   28.90         28.30

              Entry Price   Stop
First Unit    28.30         27.70
Second Unit   28.90         28.30
Third Unit    29.50         28.90

              Entry Price   Stop
First Unit    28.30         27.70
Second Unit   28.90         28.30
Third Unit    29.50         28.90
Fourth Unit   30.10         29.50
"""




# E.g. usage
if __name__ == "__main__":

  n = 1.20 # N @ $28.30 - the 55 day breakout
  standard = [28.30, 28.90, 29.50, 30.10]
  gap = [28.30, 28.90, 29.50, 30.80]

  gap_ = [28.30, 28.90, 30.80, 31.40, 35.50]

  stop = getstop(n, gap_)
  print(stop.stop)


#!/usr/bin/env python
# -*- coding: UTF8 -*-
"""https://oxfordstrat.com/coasdfASD32/uploads/2016/01/turtle-rules.pdf

Volatility - The Meaning of N
The Turtles used a concept that Richard Dennis and Bill Eckhardt called N to 
represent the underlying volatility of a particular market.

N is simply the 20-day exponential moving average of the True Range, which is now
more commonly known as the ATR. Conceptually, N represents the average range in
price movement that a particular market makes in a single day, accounting for opening
gaps. N was measured in the same points as the underlying contract.

To compute the daily true range:
  True Range = Maximum(H-L, H-PDC, PDC-L)
where:
  H - Current High
  L - Current Low
  PDC - Previous Day's Close

To compute N use the following formula:
  N = (19 * PDN + TR) / 20
where:
  PDN - Previous Day's N
  TR - Current Day's True Range

Since this formula requires a previous day's N value, you must start with a 20-day
simple average of the True Range for the initial calculation.

Dollar Volatility Adjustment
The first step in determining the position size was to determine the dollar volatility
represented by the underlying market's price volatility (defined by its N).

This sounds more complicated than it is. It is determined using the simple formula:
  Dollar Volitility = N * Dollars per Point

Volatility Adjusted Position Units
The Turtles built positions in pieces which we called Units. Units were sized so that 1
N represented 1% of the account equity.

Thus, a unit for a given market or commodity can be calculated using the following
formula:
  Unit = 1% of Account / Market Dollar Volatility
  or 
  Unit = 1% of Account / N * Dollars per Point

...

Units as a measure of Risk
The limits were:
Level       Type                            Maximum Units
1           Single Market                   4 Units
2           Closely Correlated Markets      6 Units
3           Loosely Correlated Markets      10 Units
4           Single Direction                12 Units

Single Markets - A maximum of four Units per market.

Closely Correlated Markets - For markets that were closely correlated there could be
a maximum of 6 Units in one particular direction (i.e.6 long units or 6 short units).
Closely correlated markets include: heating oil and crude oil; gold and silver; Swiss
franc and Deutschmark; TBill and Eurodollar, etc.

Loosely Correlated Markets - For loosely correlated markets, there could be a
maximum of 10 Units in one particular direction. Loosely correlated markets included:
gold and copper; silver and copper, and many grain combinations that the Turtles did
not trade because of positions limits.

Single Direction - The maximum number of total Units in one direction long or short
was 12 Units. Thus, one could theoretically have had 12 Units long and 12 Units short
at the same time.
"""
import math
import pandas as pd
from types import NoneType

from yahoo_fin.stock_info import get_data

""" ignore Pandas Future Warning """
pd.options.mode.chained_assignment = None  # default='warn'


class SingleMarketsException(Exception):
  """ Single Markets - A maximum of 4 Units per market. """


class CloselyCorrelatedMarketsException(Exception):
  """ Closely Correlated Markets - A maximum of 6 Units in one particular direction. """


class LooselyCorrelatedMarketsException(Exception):
  """ Loosely Correlated Markets - A maximum of 10 Units in one particular direction. """


class SingleDirectionException(Exception):
  """ Single Direction - A maximum of 12 Units in one direction, long or short.  """


def getn(asset: str, dataframe: pd.DataFrame = None):
  """Calculate N.

  Args:
    asset: An asset's symbol.
    dataframe: A pandas dataframe, if no query is required.
  """
  return N(asset=asset, dataframe=dataframe).n


def getunit(asset: str, n: float, account_size: float):
  """Get Unit Size.

  Args:
    asset: An asset's symbol.
    n: An asset's underlying volatility.
  """
  return Unit(asset=asset, n=n, account_size=account_size)



class N:
  """This class represents an asset's underlying volatility.

  Args:
    asset: An asset's symbol.
    dataframe: A pandas dataframe, if no query is required.
  """
  def __init__(self, asset: str, dataframe: pd.DataFrame = None):
    self.dataframe = self._dataframe(asset, dataframe)
    self.true_range = self._true_range(self.dataframe)
    self.pdn = self._pdn(self.true_range)
    self.n = self._n(self.pdn, self.true_range)


  def _dataframe(self, asset: str, dataframe: pd.DataFrame = None) -> pd.DataFrame:
    """Get the asset's pandas dataframe. 

    Args:
      asset: An asset's symbol.
      dataframe: A pandas dataframe, if no query is required.
    """
    if dataframe is None:
      dataframe = get_data(ticker=asset, interval="1d")[["low", "high", "close"]].tail(21).astype(object)
    
    else:
      dataframe = dataframe[["low", "high", "close"]].astype(object)

    dataframe["previous_close"] = dataframe["close"].shift(1)
    dataframe["true_range"] = 0
    return dataframe


  def _true_range(self, dataframe: pd.DataFrame) -> pd.Series:
    """Create and return the true range column.

    Args:
      dataframe: The asset's dataframe.
    """
    for index, row in enumerate(dataframe.iterrows()):
      # Set last previous close to 0, otherwise will NoneType error
      if isinstance(row[1]["previous_close"], NoneType):
        row[1]["previous_close"] = 0

      maximum = max(row[1]["high"] - row[1]["low"], row[1]["high"] - row[1]["previous_close"], row[1]["previous_close"] - row[1]["low"])
      dataframe.iloc[index, dataframe.columns.get_loc('true_range')] = maximum

    return dataframe["true_range"]
  

  def _pdn(self, true_range: pd.Series) -> float:
    """Get the previous day's N. 

    Args:
      true_range: The true range column. 
    """
    return sum(true_range[1:]) / 20
  

  def _n(self, pdn: float, true_range: pd.Series) -> float:
    """Get 'N'.
    
    Args:
      pdn: The previous day's 'N'.
      true_range: The true range column. 
    """
    return (19 * pdn + true_range[-1]) / 20
  
  

class Unit:
  """This class represents an asset's units size.

  Args:
    asset: An asset's symbol.
    N: An asset's underlying volatility.
  """
  def __init__(self, asset: str, n: float, account_size: float):
    self.dollar_volatility = self._dollar_volatility(asset, n)
    self.unit = self._unit_size(self.dollar_volatility, account_size)


  def _dollar_volatility(self, asset: str, n: float) -> float:
    """Get the market dollar volatility of an asset.

    Args:
      asset: An asset's symbol.
    """
    # get the live price
    dataframe = get_data(ticker=asset, end_date=pd.Timestamp.today() + pd.DateOffset(10))
    pps = dataframe.close[-1]

    # multiply N and the live price
    return n * pps
  

  def _unit_size(self, dollar_volatility, account_size):
    """Get the volatility adjusted unit size.
    
    Args:
      dollar_volatility: The market dollar volatility of an asset.
    """
    return math.trunc((0.01 * account_size) / dollar_volatility)